@@ -105,11 +105,17 @@ module Control.Concurrent.Async (
     waitEither_,
     waitBoth,
 
+    -- ** Waiting for multiple 'Async's in STM
+    waitAnySTM, waitAnyCatchSTM,
+    waitEitherSTM, waitEitherCatchSTM,
+    waitEitherSTM_,
+    waitBothSTM,
+
     -- ** Linking
     link, link2,
 
     -- * Convenient utilities
-    race, race_, concurrently, mapConcurrently,
+    race, race_, concurrently, mapConcurrently, forConcurrently,
     Concurrently(..),
 
   ) where
@@ -122,7 +128,14 @@ import Prelude hiding (catch)
 #endif
 import Control.Monad
 import Control.Applicative
+#if !MIN_VERSION_base(4,8,0)
+import Data.Monoid (Monoid(mempty,mappend))
 import Data.Traversable
+#endif
+#if MIN_VERSION_base(4,9,0)
+import Data.Semigroup (Semigroup((<>)))
+#endif
+
 
 import GHC.Exts
 import GHC.IO hiding (finally, onException)
@@ -316,9 +329,15 @@ cancelWith (Async t _) e = throwTo t e
 -- If multiple 'Async's complete or have completed, then the value
 -- returned corresponds to the first completed 'Async' in the list.
 --
+{-# INLINE waitAnyCatch #-}
 waitAnyCatch :: [Async a] -> IO (Async a, Either SomeException a)
-waitAnyCatch asyncs =
-  atomically $
+waitAnyCatch = atomically . waitAnyCatchSTM
+
+-- | A version of 'waitAnyCatch' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitAnyCatchSTM :: [Async a] -> STM (Async a, Either SomeException a)
+waitAnyCatchSTM asyncs =
     foldr orElse retry $
       map (\a -> do r <- waitCatchSTM a; return (a, r)) asyncs
 
@@ -336,9 +355,15 @@ waitAnyCatchCancel asyncs =
 -- If multiple 'Async's complete or have completed, then the value
 -- returned corresponds to the first completed 'Async' in the list.
 --
+{-# INLINE waitAny #-}
 waitAny :: [Async a] -> IO (Async a, a)
-waitAny asyncs =
-  atomically $
+waitAny = atomically . waitAnySTM
+
+-- | A version of 'waitAny' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitAnySTM :: [Async a] -> STM (Async a, a)
+waitAnySTM asyncs =
     foldr orElse retry $
       map (\a -> do r <- waitSTM a; return (a, r)) asyncs
 
@@ -350,11 +375,19 @@ waitAnyCancel asyncs =
   waitAny asyncs `finally` mapM_ cancel asyncs
 
 -- | Wait for the first of two @Async@s to finish.
+{-# INLINE waitEitherCatch #-}
 waitEitherCatch :: Async a -> Async b
                 -> IO (Either (Either SomeException a)
                               (Either SomeException b))
-waitEitherCatch left right =
-  atomically $
+waitEitherCatch left right = atomically (waitEitherCatchSTM left right)
+
+-- | A version of 'waitEitherCatch' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitEitherCatchSTM :: Async a -> Async b
+                -> STM (Either (Either SomeException a)
+                               (Either SomeException b))
+waitEitherCatchSTM left right =
     (Left  <$> waitCatchSTM left)
       `orElse`
     (Right <$> waitCatchSTM right)
@@ -372,18 +405,30 @@ waitEitherCatchCancel left right =
 -- that finished first raised an exception, then the exception is
 -- re-thrown by 'waitEither'.
 --
+{-# INLINE waitEither #-}
 waitEither :: Async a -> Async b -> IO (Either a b)
-waitEither left right =
-  atomically $
+waitEither left right = atomically (waitEitherSTM left right)
+
+-- | A version of 'waitEither' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitEitherSTM :: Async a -> Async b -> STM (Either a b)
+waitEitherSTM left right =
     (Left  <$> waitSTM left)
       `orElse`
     (Right <$> waitSTM right)
 
 -- | Like 'waitEither', but the result is ignored.
 --
+{-# INLINE waitEither_ #-}
 waitEither_ :: Async a -> Async b -> IO ()
-waitEither_ left right =
-  atomically $
+waitEither_ left right = atomically (waitEitherSTM_ left right)
+
+-- | A version of 'waitEither_' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitEitherSTM_:: Async a -> Async b -> STM ()
+waitEitherSTM_ left right =
     (void $ waitSTM left)
       `orElse`
     (void $ waitSTM right)
@@ -399,9 +444,15 @@ waitEitherCancel left right =
 -- an exception before they have both finished, then the exception is
 -- re-thrown by 'waitBoth'.
 --
+{-# INLINE waitBoth #-}
 waitBoth :: Async a -> Async b -> IO (a,b)
-waitBoth left right =
-  atomically $ do
+waitBoth left right = atomically (waitBothSTM left right)
+
+-- | A version of 'waitBoth' that can be used inside an STM transaction.
+--
+-- @since 2.1.0
+waitBothSTM :: Async a -> Async b -> STM (a,b)
+waitBothSTM left right = do
     a <- waitSTM left
            `orElse`
          (waitSTM right >> retry)
@@ -519,7 +570,9 @@ concurrently' left right collect = do
                              `catchAll` (putMVar done . Left)
         rid <- forkIO $ restore (right >>= putMVar done . Right . Right)
                              `catchAll` (putMVar done . Left)
-        let stop = killThread lid >> killThread rid
+        let stop = killThread rid >> killThread lid
+                   -- kill right before left, to match the semantics of
+                   -- the version using withAsync. (#27)
         r <- restore (collect done) `onException` stop
         stop
         return r
@@ -538,6 +591,14 @@ concurrently' left right collect = do
 mapConcurrently :: Traversable t => (a -> IO b) -> t a -> IO (t b)
 mapConcurrently f = runConcurrently . traverse (Concurrently . f)
 
+-- | `forConcurrently` is `mapConcurrently` with its arguments flipped
+--
+-- > pages <- forConcurrently ["url1", "url2", "url3"] $ \url -> getURL url
+--
+-- @since 2.1.0
+forConcurrently :: Traversable t => t a -> (a -> IO b)-> IO (t b)
+forConcurrently = flip mapConcurrently
+
 -- -----------------------------------------------------------------------------
 
 -- | A value of type @Concurrently a@ is an @IO@ operation that can be
@@ -571,10 +632,23 @@ instance Alternative Concurrently where
   Concurrently as <|> Concurrently bs =
     Concurrently $ either id id <$> race as bs
 
-instance Monad Concurrently where
-  return = pure
-  Concurrently a >>= f =
-    Concurrently $ a >>= runConcurrently . f
+#if MIN_VERSION_base(4,9,0)
+-- | Only defined by @async@ for @base >= 4.9@
+--
+-- @since 2.1.0
+instance Semigroup a => Semigroup (Concurrently a) where
+  (<>) = liftA2 (<>)
+
+-- | @since 2.1.0
+instance (Semigroup a, Monoid a) => Monoid (Concurrently a) where
+  mempty = pure mempty
+  mappend = (<>)
+#else
+-- | @since 2.1.0
+instance Monoid a => Monoid (Concurrently a) where
+  mempty = pure mempty
+  mappend = liftA2 mappend
+#endif
 
 -- ----------------------------------------------------------------------------
 
